/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/




// CHANGES

#include "ABKCommon/abkcommon.h"
#include "ABKCommon/abkversion.h"
#include "DB/DB.h"
#include "Capo/capoPlacer.h"
#include "RBPlFromDB/RBPlaceFromDB.h"
#include "NetTopology/hgWDimsFromDB.h"
#include "FixOr/greedyOrientOpt.h"
#include "RowIroning/rowIroning.h"

int main(int argc, const char *argv[])
{
  NoParams         noParams       (argc,argv);
  BoolParam        helpRequest    ("help",argc,argv);
  BoolParam        helpRequest1   ("h",   argc,argv);
  BoolParam        saveDef        ("save",argc,argv);
  BoolParam        doCellFlipping ("flipping",argc,argv);
  BoolParam        doRowIroning   ("rowIroning",argc,argv);

  Verbosity	   verb(argc, argv);

  PRINT_VERSION_INFO
  cout << CmdLine(argc,argv);

  CapoParameters capoParams(argc,argv);

  RowIroningParameters RIparams(argc, argv);
  if (noParams.found() || helpRequest.found() || helpRequest1.found())
  {
     cerr << "  use '-help' or '-f filename.aux' " << endl;
     cerr << " -flipint    "<<endl;
     cerr << " -rowIroning "<<endl;
     cerr << RIparams;

     return 0;
  }

  cout << capoParams << endl;
  if (doRowIroning.found()) cout << RIparams << endl;

  StringParam 	 auxFileName 	("f", 	argc,argv);
  abkfatal(auxFileName.found(),"Usage: prog -f filename.aux <more params>");

  DB db(auxFileName);
  cerr<<"Design Name: "<<db.getDesignName()<<endl;


  RBPlaceParams rbParam(argc,argv);
  RBPlaceFromDB rbplace(db, rbParam);   

//const HGraphWDimensions &hgraph = rbplace.getHGraph();

  Timer totalPlTime;
  Timer capoTimer;
  CapoPlacer capo(rbplace, capoParams);
  capoTimer.stop();

  db.setPlaceAndOrient(rbplace);
       
  cout << "After Capo  "<<endl;
  cout << "  Capo Runtime:             "<<capoTimer.getUserTime()<<endl;

  cout << "  DB   Est Pin-to-Pin HPWL: "<<
	db.getNetlist().evalHalfPerimCost(db.spatial.locations,
				  db.spatial.orientations)<<endl<<endl;

  cout << "  RBPl Est            HPWL: "<<rbplace.evalHPWL(false)<<endl;
  cout << "  RBPl Est Pin-to-Pin HPWL: "<<rbplace.evalHPWL(true)<<endl;
  cout << "  RBPl Est             WWL: "<<rbplace.evalWeightedWL(false)<<endl;
  cout << "  RBPl Est Pin-to-Pin  WWL: "<<rbplace.evalWeightedWL(true)<<endl<<endl;


  if(doCellFlipping.found() && !doRowIroning.found())
  { 
      OrientOptParameters flipParams;

      Timer flipTimer;
      GreedyOrientOpt flipper(db,flipParams);
      flipTimer.stop();

      db<<flipper;

      cout<<"After Greedy Flipping"<<endl;

      cout << "  Flip Runtime: "<<flipTimer.getUserTime()<<endl;
      cout << "  Total Cells flipped: "<<flipper.getNumFlips()<<endl;
      cout << "  DB   Est Pin-to-Pin HPWL: "<<
	db.getNetlist().evalHalfPerimCost(db.spatial.locations,
				  db.spatial.orientations)<<endl<<endl;
  }

  if(!doCellFlipping.found() && doRowIroning.found())
  {
     Timer RITimer;
     ironRows(rbplace, RIparams);
     RITimer.stop();
     db.setPlaceAndOrient(rbplace);

     cout << "  RI Runtime: "<<RITimer.getUserTime()<<endl;
     cout << "  DB   Est Pin-to-Pin HPWL: "<<
	db.getNetlist().evalHalfPerimCost(db.spatial.locations,
				  db.spatial.orientations)<<endl<<endl;
  }

  if(doCellFlipping.found() && doRowIroning.found())
  {
     unsigned numPasses = RIparams.numPasses;
     RIparams.numPasses = 1;
     RIparams.verb = "0_0_0";

     for(unsigned p = 0; p < numPasses; p++)
     {
         cout<<" Ironing/Flipping Pass: "<<p<<endl;

         Timer ironTime;
         ironRows(rbplace, RIparams);
         ironTime.stop();
         db.setPlaceAndOrient(rbplace);

         cout <<"   RI Runtime:  " << ironTime.getUserTime()<<endl;
         cout << "  DB   Est Pin-to-Pin HPWL: "<<
	    db.getNetlist().evalHalfPerimCost(db.spatial.locations,
				  db.spatial.orientations)<<endl<<endl;

         OrientOptParameters flipParams;
         Timer flipTimer;
         GreedyOrientOpt flipper(db,flipParams);
         flipTimer.stop();

         db<<flipper;
         cout << "  Flip Runtime: "<<flipTimer.getUserTime()<<endl;
         cout << "  DB   Est Pin-to-Pin HPWL: "<<
	    db.getNetlist().evalHalfPerimCost(db.spatial.locations,
				  db.spatial.orientations)<<endl<<endl;

         rbplace.resetPlacementWOri(
		PlacementWOrient(db.spatial.locations,
		   	      db.spatial.orientations));

     }
  }

  totalPlTime.stop();

  cout << "Total runtime "<<totalPlTime.getUserTime()<<endl;
  cout << "FINAL  DB   Est Pin-to-Pin HPWL: "<<
	db.getNetlist().evalHalfPerimCost(db.spatial.locations,
				  db.spatial.orientations)<<endl<<endl;

  if (saveDef.found()) db.saveDEF("out.def");


  return 0;
}
