/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/




#include "rowIroning.h"

#include <algorithm>
#include <iostream.h>
#include "HGraphWDims/hgWDims.h"
#include "RBPlace/RBPlacement.h"
#include "RBPlace/pinLocCalc.h"
#include "SmallPlacement/smallPlaceProb.h"
#include "SmallPlacers/sRowBBPlacer.h"

void ironRows(RBPlacement& rbplace,
              RowIroningParameters params)
{
 Timer totalTime;

  const HGraphWDimensions& hgWDims = rbplace.getHGraph();

  // check that populated ?
  // check params

  abkfatal(params.windowSize>1 && params.windowSize<20,
            " Row ironing window size must be >1 and <20");
//abkfatal(params.overlap>0 && params.overlap<params.windowSize,
//          " Row ironing overlap must be >0 and smaller than window size\n");
  abkfatal(                    params.overlap<params.windowSize,
            " Row ironing overlap must be        smaller than window size\n");
  abkfatal(params.numPasses<20, "Row ironing cannot have 20+ passes");

 for(unsigned pass=0; pass!=params.numPasses; pass++)
 {
   Timer passTime;
   itRBPCoreRow rIt;

   for (rIt=rbplace.coreRowsBegin(); rIt!=rbplace.coreRowsEnd(); rIt++)
   {
     for (itRBPSubRow sIt=rIt->subRowsBegin(); sIt!=rIt->subRowsEnd(); sIt++)
     {
        const RBPSubRow& subrow=*sIt;

        SmallPlacementRow smallRow;
        smallRow.siteInterval=subrow.getSpacing();
        smallRow.xMax =smallRow.xMin=subrow.getXStart();
        smallRow.yMax =smallRow.yMin=subrow.getYCoord();
        smallRow.yMax+=subrow.getHeight();
//      double   xEnd=subrow.getXEnd();
        unsigned firstCell=0, totalCells=subrow.getNumCells();
        if (totalCells<2) continue;

        while (firstCell+params.windowSize<totalCells)
        {
           unsigned cellAfterLast=min(firstCell+params.windowSize,totalCells);
           vector<unsigned> movables(cellAfterLast-firstCell);
           std::copy(subrow.cellIdsBegin()+firstCell,
 	        subrow.cellIdsBegin()+cellAfterLast, movables.begin());
           if (firstCell>0) smallRow.xMin=rbplace[subrow[firstCell]].x;
           if (cellAfterLast!=totalCells) 
           smallRow.xMax=rbplace[subrow[cellAfterLast-1]].x+subrow.getSpacing();
           else  // end of the row
           {
             smallRow.xMax=subrow.getXEnd();
             double totalCellWidth=0.0;
             for(unsigned k=0; k!=movables.size(); k++)
              totalCellWidth+= hgWDims.getNodeWidth(movables[k]);
             double aveWhiteSpacePerCell
                 =(smallRow.xMax-smallRow.xMin-totalCellWidth)/movables.size();
             smallRow.xMax+=aveWhiteSpacePerCell;
           }

	   if(movables.size() < 2) continue;

           SmallPlacementProblem smallProblem(hgWDims, rbplace,
            rbplace.getFixed(), movables, smallRow);
            
           SmallPlacementSolution& solution=smallProblem.getSolnBuffer()[0];
           solution.populated=true;
           for(unsigned cell=0; cell!=movables.size(); cell++)
           {
              solution.placement[cell]   =rbplace[movables[cell]];
              solution.orientations[cell]=rbplace.getOrient(movables[cell]);
           }

           BranchSmallPlacer placer(smallProblem,params.smplParams);

	   //get permutation coresponding to the produced placement 
	    vector<unsigned> perm;
	    smallProblem.getBestSoln().setPermutation(perm);  

	   rbplace.permuteCellsInSR(const_cast<RBPSubRow&>(subrow), 
					firstCell, perm);

           firstCell+= (params.windowSize - params.overlap);
        }
     }
   }
   passTime.stop();
   if (params.verb.forMajStats || params.verb.forSysRes )
   {
      cout << " Pass " << pass << " took : " << passTime << endl;
      cout << "  Pin-to-Pin HPWL after pass: "
		<<rbplace.evalHPWL(true)<<endl;
   }
 }
 totalTime.stop();
 if (params.verb.forMajStats || params.verb.forSysRes )
      cout << " Row ironing took : " << totalTime << endl;
}

RowIroningParameters::RowIroningParameters(int argc, const char *argv[])
 : numPasses(1), windowSize(7), overlap(2), smplParams(argc,argv),
   verb(argc,argv)
{
  NoParams         noParams       (argc,argv);
  BoolParam        helpRequest    ("help",argc,argv);
  BoolParam        helpRequest1   ("h",   argc,argv);
  UnsignedParam numPasses_ ("ironPasses",argc,argv);
  UnsignedParam windowSize_("ironWindow",argc,argv);
  UnsignedParam overlap_   ("ironOverlap",argc,argv);

  if (noParams.found() || helpRequest.found() || helpRequest1.found())
  {
     cout << " Row Ironing Parameters: " 
          << "\n   -ironPasses <unsigned> "
          << "\n   -ironWindow <unsigned> "
          << "\n   -ironOverlap <unsigned>" << endl;
     return;
  }
  if (numPasses_. found())  numPasses =numPasses_;
  if (windowSize_.found())  windowSize=windowSize_;
  if (overlap_.   found())  overlap   =overlap_;
}

ostream& operator<<(ostream& out, const RowIroningParameters& params)
{
   out << " Row ironing paramerers : "
       << "\n   Num. passes    : " << params.numPasses 
       << "\n   Window size    : " << params.windowSize
       << "\n   Overlap        : " << params.overlap
       << "\n   Verbosity      : " << params.verb << endl;
  return out;
}
